#-*- coding: utf-8 -*-
# pylint: disable=protected-access, too-few-public-methods, no-member

"""This module is used by the `model <#module-half_orm.model>`_ module
to generate the classes that manipulate the data in your database
with the `Model.get_relation_class <#half_orm.model.Model.get_relation_class>`_
method.


Example:
    >>> from half_orm.model import Model
    >>> model = Model('halftest')
    >>> class Person(model.get_relation_class('actor.person')):
    >>>     # your code goes here
"""

"""
Main methods provided by the class Relation:
- insert: inserts a tuple into the pg table.
- select: returns a generator of the elements of the set defined by
  the constraint on the Relation object. The elements are dictionaries with the
  keys corresponding to the selected columns names in the relation.
  The result is affected by the methods: distinct, order_by, limit and offset
  (see bellow).
- update: updates the set defined by the constraint on the Relation object
  with the values passed as arguments.
- delete: deletes from the relation the set of elements defined by the constraint
  on the Relation object.
- get: returns the unique element defined by the constraint on the Relation object.
  the element returned if of the type of the Relation object.
- count: returns the number of elements in the set defined by the constraint on the
  Relation object.

The following methods can be chained on the object before a select.

- distinct: ensures that there are no duplicates on the select result.
- order_by: sets the order of the select result.
- limit: limits the number of elements returned by the select method.
- offset: sets the offset for the select method.

"""

from functools import wraps
from collections import OrderedDict
from uuid import UUID
from typing import Generator, List
from datetime import date, datetime, time, timedelta
import json
import sys
import psycopg2


import yaml

from half_orm import relation_errors
from half_orm.transaction import Transaction
from half_orm.field import Field

class _SetOperators:
    """_SetOperators class stores the set operations made on the Relation class objects

    - __operator is one of {'or', 'and', 'sub', 'neg'}
    - __right is a Relation object. It can be None if the operator is 'neg'.
    """
    def __init__(self, left, operator=None, right=None):
        self.__left = left
        self.__operator = operator
        self.__right = right

    @property
    def operator(self):
        """Property returning the __operator value."""
        return self.__operator
    @operator.setter
    def operator(self, operator):
        """Set operator setter."""
        self.__operator = operator

    @property
    def left(self):
        """Returns the left object of the set operation."""
        return self.__left
    @left.setter
    def left(self, left):
        """left operand (relation) setter."""
        self.__left = left

    @property
    def right(self):
        """Property returning the right operand (relation)."""
        return self.__right
    @right.setter
    def right(self, right):
        """right operand (relation) setter."""
        self.__right = right

    def __repr__(self):
        return f"{self.__operator} {self.__right and self.__right._fqrn or None}"

class Relation:
    """Used as a base class for the classes generated by
    `Model.get_relation_class <#half_orm.model.Model.get_relation_class>`_.

    Args:
        **kwargs: the arguments names must correspond to the columns names of the relation.

    Raises:
        UnknownAttributeError: If the name of an argument doesn't match a column name in the
            relation considered.

    Examples:
        You can generate a class for any relation in your database:
            >>> from half_orm.model import Model
            >>> model = Model('halftest')
            >>> class Person(model.get_relation_class('actor.person')):
            >>>     # your code

        To define a set of data in your relation at instantiation:
            >>> gaston = Person(last_name='Lagaffe', first_name='Gaston')
            >>> all_names_starting_with_la = Person(last_name=('ilike', 'la%'))

        Or to constrain an instantiated object via its\
            `Fields <#half_orm.field.Field>`_:
            >>> person = Person()
            >>> person.birth_date = ('>', '1970-01-01')

        Raises an `UnknownAttributeError <#half_orm.relation_errors.UnknownAttributeError>`_:
            >>> Person(lost_name='Lagaffe')
            [...]UnknownAttributeError: ERROR! Unknown attribute: {'lost_name'}.
    """

#### THE following METHODS are included in Relation class according to
#### relation type (Table or View). See TABLE_INTERFACE and VIEW_INTERFACE.

def __init__(self, **kwargs):
    _fqrn = ""
    """The arguments names must correspond to the columns names of the relation.
    """
    self._fields = {}
    self._pkey = {}
    self._fkeys = OrderedDict()
    self._fkeys_prop = []
    self._fkeys_attr = set()
    self.__only = False
    self.__neg = False
    self.__set_fields()
    self.__set_fkeys()
    self._joined_to = {}
    self.__query = ""
    self.__query_type = None
    self.__sql_query = []
    self.__sql_values = []
    self.__set_operators = _SetOperators(self)
    self.__select_params = {}
    self.__id_cast = None
    self.__cursor = self._model._connection.cursor()
    self.__cons_fields = []
    self.__mogrify = False
    self._is_singleton = False
    kwk_ = set(kwargs.keys())
    if kwk_.intersection(self._fields.keys()) != kwk_:
        raise relation_errors.UnknownAttributeError(str(kwk_.difference(self._fields.keys())))
    _ = {self.__dict__[field_name]._set(value)
         for field_name, value in kwargs.items() if value is not None}
    self.__isfrozen = True

def insert(self, *args) -> '[dict]':
    """Insert a new tuple into the Relation.

    Returns:
        [dict]: A singleton containing the data inserted.

    Example:
        >>> gaston = Person(last_name='Lagaffe', first_name='Gaston', birth_date='1970-01-01').insert()
        >>> print(gaston)
        [{'id': 1772, 'first_name': 'Gaston', 'last_name': 'Lagaffe', 'birth_date': datetime.date(1970, 1, 1)}]

    Note:
        It is not possible to insert more than one row with the insert method
    """
    query_template = "insert into {} ({}) values ({})"
    self.__query_type = 'insert'
    fields_names, values, fk_fields, fk_query, fk_values = self.__what_to_insert()
    what_to_insert = ["%s" for _ in range(len(values))]
    if fk_fields:
        fields_names += fk_fields
        what_to_insert += fk_query
        values += fk_values
    query = self.__add_returning(
        query_template.format(self._qrn, ", ".join(fields_names), ", ".join(what_to_insert)), *args)
    self.__execute(query, tuple(values))
    return [dict(elt) for elt in self.__cursor.fetchall()]

def select(self, *args):
    """Gets the set of values correponding to the constraint attached to the object.
    This method is a generator.

    Arguments:
        *args: the fields names of the returned attributes. If omitted,
            all the fields are returned.

    Yields:
        the result of the query as a dictionary.

    Example:
        >>> for person in Person(last_name=('like', 'La%')).select('id'):
        >>>     print(person)
        {'id': 1772}
    """
    query, values = self._prep_select(*args)
    try:
        self.__execute(query, values)
    except Exception as err:
        sys.stderr.write(f"QUERY: {query}\nVALUES: {values}\n")
        raise err
    for elt in self.__cursor:
        yield dict(elt)

def get(self, *args: List[str]) -> Relation:
    """The get method allows you to fetch a singleton from the database.
    It garantees that the constraint references one and only one tuple.

    Args:
        args (List[str]): list of fields names.\
        If ommitted, all the values of the row retreived from the database\
        are set for the self object.\
        Otherwise, only the values listed in the `args` parameter are set.

    Returns:
        Relation: the object retreived from the database.

    Raises:
        ExpectedOneError: an exception is raised if no or more than one element is found.

    Example:
        >>> gaston = Person(last_name='Lagaffe', first_name='Gaston').get()
        >>> type(gaston) is Person
        True
        >>> gaston.id
        (int4) NOT NULL (id = 1772)
    """
    _count = len(self)
    if _count != 1:
        raise relation_errors.ExpectedOneError(self, _count)
    self._is_singleton = True
    ret = self(**(next(self.select(*args))))
    ret._is_singleton = True
    return ret

def update(self, *args, update_all=False, **kwargs):
    """
    kwargs represents the values to be updated {[field name:value]}
    The object self must be set unless update_all is True.
    The constraints of the relations are updated with kwargs.
    """
    if not (self.is_set() or update_all):
        raise RuntimeError(
            f'Attempt to update all rows of {self.__class__.__name__}'
            ' without update_all being set to True!')

    update_args = dict(kwargs)
    for key, value in kwargs.items():
        # None values are first removed
        if value is None:
            update_args.pop(key)
    if not update_args:
        return []# no new value update. Should we raise an error here?

    query_template = "update {} set {} {}"
    what, where, values = self.__update_args(**update_args)
    _, _, fk_fields, fk_query, fk_values = self.__what_to_insert()
    if fk_fields:
        fk_where = " and ".join([f"({a}) in ({b})" for a, b in zip(fk_fields, fk_query)])
        where = f"{where} and {fk_where}"
        values += fk_values
    query = self.__add_returning(query_template.format(self._qrn, what, where), *args)
    self.__execute(query, tuple(values))
    for field_name, value in update_args.items():
        self._fields[field_name]._set(value)
    return [dict(elt) for elt in self.__cursor.fetchall()]

def delete(self, *args, delete_all=False):
    """Removes a set of tuples from the relation.
    To empty the relation, delete_all must be set to True.
    """
    if not (self.is_set() or delete_all):
        raise RuntimeError(
            f'Attempt to delete all rows from {self.__class__.__name__}'
            ' without delete_all being set to True!')
    query_template = "delete from {} {}"
    _, values = self.__get_query(query_template)
    self.__query_type = 'delete'
    _, where, _ = self.__where_args()
    _, _, fk_fields, fk_query, fk_values = self.__what_to_insert()
    where = f" where {where}"
    if where == "(1 = 1)" and not delete_all:
        raise RuntimeError
    if fk_fields:
        fk_where = " and ".join([f"({a}) in ({b})" for a, b in zip(fk_fields, fk_query)])
        where = f"{where} and {fk_where}"
        values += fk_values
    query = query_template.format(self._qrn, where)
    query = self.__add_returning(query, *args)
    self.__execute(query, tuple(values))
    return [dict(elt) for elt in self.__cursor.fetchall()]

def __add_returning(self, query, *args) -> str:
    "Adds the SQL returning clause to the query"
    returning = '*'
    if args:
        returning = ','.join(args)
    return f'{query} returning {returning}'

def _unfreeze(self):
    "Allow to add attributs to a relation"
    self.__isfrozen = False

def _freeze(self):
    "set __isfrozen to True."
    self.__isfrozen = True

def __setattr__(self, key, value):
    """Sets an attribute as long as __isfrozen is False

    The foreign keys properties are not detected by hasattr
    hence the line `_ = self.__dict__[key]` to double check if
    the attribute is really present.
    """
    if not hasattr(self, '__isfrozen'):
        object.__setattr__(self, '__isfrozen', False)
    if self.__isfrozen and not hasattr(self, key):
        raise relation_errors.IsFrozenError(self.__class__, key)
    if self.__dict__.get(key) and isinstance(self.__dict__[key], Field):
        self.__dict__[key]._set(value)
        return
    object.__setattr__(self, key, value)

def __execute(self, query, values):
    try:
        if self.__mogrify:
            print(self.__cursor.mogrify(query, values).decode('utf-8'))
        return self.__cursor.execute(query, values)
    except (psycopg2.OperationalError, psycopg2.InterfaceError):
        self._model.ping()
        self.__cursor = self._model._connection.cursor()
        return self.__cursor.execute(query, values)

@property
def id_(self):
    """Return the __id_cast or the id of the relation.
    """
    return self.__id_cast or id(self)

@property
def only(self):
    "Returns the value of self.__only"
    return self.__only
@only.setter
def only(self, value):
    """Set the value of self.__only. Restrict the values of a query to
    the elements of the relation (no inherited values).
    """
    if not value in {True, False}:
        raise ValueError(f'{value} is not a bool!')
    self.__only = value

def __set_fields(self):
    """Initialise the fields of the relation."""
    _fields_metadata = self._model._fields_metadata(self._t_fqrn)

    for field_name, f_metadata in _fields_metadata.items():
        field = Field(field_name, self, f_metadata)
        self._fields[field_name] = field
        self.__setattr__(field_name, field)
        if field._is_part_of_pk():
            self._pkey[field_name] = field

def __set_fkeys(self):
    """Initialisation of the foreign keys of the relation"""
    #pylint: disable=import-outside-toplevel
    from half_orm.fkey import FKey

    _fkeys_metadata = self._model._fkeys_metadata(self._t_fqrn)
    for fkeyname, f_metadata in _fkeys_metadata.items():
        self._fkeys[fkeyname] = FKey(fkeyname, self, *f_metadata)
    #TODO: Remove in 0.8 release
    if not self.__fkeys_properties:
        self._set_fkeys_properties()
        self.__fkeys_properties = True
    if hasattr(self, 'Fkeys'):
        for key, value in self.Fkeys.items():
            try:
                if key != '': # we skip empty keys
                    setattr(self.__class__.__base__, key, self._fkeys[value])
                    self._fkeys_attr.add(key)
            except KeyError as exp:
                raise relation_errors.WrongFkeyError(self, value) from exp

def _set_fkeys_properties(self):
    """Property generator for fkeys.
    @args is a list of tuples (property_name, fkey_name)

    Will be removed in 0.8 release.
    """
    fkp = __import__(self.__module__, globals(), locals(), ['FKEYS_PROPERTIES', 'FKEYS'], 0)
    if hasattr(fkp, 'FKEYS_PROPERTIES'):
        sys.stderr.write(
            'WARNING! Deprecation (FKEYS_PROPERTIES will be removed in half_orm 0.8 release):\n'
            f'Please replace {self.__module__}.FKEYS_PROPERTIES variable with the {self.__class__.__name__}.Fkeys class attribute.\n')
        for prop in fkp.FKEYS_PROPERTIES:
            self._set_fkey_property(*prop)
    if hasattr(fkp, 'FKEYS'):
        sys.stderr.write(
            'WARNING! Deprecation (FKEYS will be removed in half_orm 0.8 release):\n'
            f'Please replace {self.__module__}.FKEYS variable with the {self.__class__.__name__}.Fkeys class attribute.\n')
        for prop in fkp.FKEYS:
            self._set_fkey_property(*prop)

def _set_fkey_property(self, property_name, fkey_name, _cast=None):
    """Sets the property with property_name on the foreign key."""
    if property_name == '':
        # Do nothing
        return
    def fget(self):
        "getter"
        return self._fkeys[fkey_name](__cast__=_cast)
    def fset(self, value):
        "setter"
        try:
            if value.is_set():
                self._fkeys[fkey_name].set(value)
        except KeyError as err:
            sys.stderr.write(
                f'ERR {err}\nFKeys for {self.__class__.__name__} are: {self._fkeys.keys()}\n')
            raise err
    if self.__dict__.get(property_name):
        raise relation_errors.DuplicateAttributeError(
            f"ERROR: Can't set '{property_name}' as a FKEY property in {self.__class__}!")
    self._fkeys_prop.append(property_name)
    setattr(self.__class__, property_name, property(fget=fget, fset=fset))

def group_by(self, yml_directive):
    """Returns an aggregation of the data according to the yml directive
    description.
    """
    def inner_group_by(data, directive, grouped_data, gdata=None):
        """recursive fonction to actually group the data in grouped_data."""
        deja_vu_key = set()
        if gdata is None:
            gdata = grouped_data
        if isinstance(directive, list):
            directive = directive[0]
        keys = set(directive)
        for elt in data:
            res_elt = {}
            for key in keys.intersection(self._fields.keys()):
                deja_vu_key.add(directive[key])
                try:
                    res_elt.update({directive[key]:elt[key]})
                except KeyError as exc:
                    raise relation_errors.UnknownAttributeError(key) from exc
            if isinstance(gdata, list):
                different = None
                for selt in gdata:
                    different = True
                    for key in deja_vu_key:
                        different = selt[key] != res_elt[key]
                        if different:
                            break
                    if not different:
                        break
                if not gdata or different:
                    gdata.append(res_elt)
            else:
                gdata.update(res_elt)
            for group_name in keys.difference(
                    keys.intersection(self._fields.keys())):
                type_directive = type(directive[group_name])
                suite = None
                if not gdata:
                    gdata[group_name] = type_directive()
                    suite = gdata[group_name]
                elif isinstance(gdata, list):
                    suite = None
                    for selt in gdata:
                        different = True
                        for skey in deja_vu_key:
                            different = selt[skey] != res_elt[skey]
                            if different:
                                break
                        if not different:
                            if selt.get(group_name) is None:
                                selt[group_name] = type_directive()
                            suite = selt[group_name]
                            break
                    if suite is None:
                        gdata.append(res_elt)
                elif gdata.get(group_name) is None:
                    #TODO: Raise ExpectedOneError if necessary
                    gdata[group_name] = type_directive()
                    suite = gdata[group_name]
                else:
                    suite = gdata[group_name]
                inner_group_by(
                    [elt], directive[group_name], suite, None)

    grouped_data = {}
    data = list(self.select())
    directive = yaml.safe_load(yml_directive)
    inner_group_by(data, directive, grouped_data)
    return grouped_data

def to_json(self, yml_directive=None, res_field_name='elements', **kwargs):
    """Returns a JSON representation of the set returned by the select query.
    if kwargs, returns {res_field_name: [list of elements]}.update(kwargs)
    """

    def handler(obj):
        """Replacement of default handler for json.dumps."""
        if hasattr(obj, 'isoformat'):
            return str(obj.isoformat())
        if isinstance(obj, UUID):
            return str(obj)
        if isinstance(obj, timedelta):
            return obj.total_seconds()
        raise TypeError(
            f'Object of type {type(obj)} with value of {repr(obj)} is not JSON serializable')

    if yml_directive:
        res = self.group_by(yml_directive)
    else:
        res = [elt for elt in self.select()]
    if kwargs:
        res = {res_field_name: res}
        res.update(kwargs)
    return json.dumps(res, default=handler)

def to_dict(self):
    """Returns a dictionary containing only the values of the fields
    that are set."""
    return {key:field.value for key, field in self._fields.items() if field.is_set()}

def _to_dict_val_comp(self):
    """Returns a dictionary containing the values and comparators of the fields
    that are set."""
    return {key:(field._comp(), field.value) for key, field in
            self._fields.items() if field.is_set()}

def __repr__(self):

    fkeys_usage = """To use the foreign keys as direct attributes of the class, copy/paste the Fkeys bellow in
your code as a class attribute and replace the empty string(s) key(s) with the alias you
want to use. The aliases must be unique and different from any of the column names. Empty
string keys are ignored.

Fkeys = {"""

    rel_kind = self.__kind
    ret = []
    ret.append(f"__RCLS: {self.__class__}")
    ret.append(
        "This class allows you to manipulate the data in the PG relation:")
    ret.append(f"{rel_kind.upper()}: {self._fqrn}")
    if self.__metadata['description']:
        ret.append(f"DESCRIPTION:\n{self.__metadata['description']}")
    ret.append('FIELDS:')
    mx_fld_n_len = 0
    for field_name in self._fields.keys():
        if len(field_name) > mx_fld_n_len:
            mx_fld_n_len = len(field_name)
    for field_name, field in self._fields.items():
        ret.append(f"- {field_name}:{' ' * (mx_fld_n_len + 1 - len(field_name))}{repr(field)}")
    ret.append('')
    pkey = self._model._pkey_constraint(self._t_fqrn)
    if pkey:
        ret.append(f"PRIMARY KEY ({', '.join(pkey)})")
    for uniq in self._model._unique_constraints_list(self._t_fqrn):
        ret.append(f"UNIQUE CONSTRAINT ({', '.join(uniq)})")
    if self._fkeys.keys():
        plur = 'S' if len(self._fkeys) > 1 else ''
        ret.append(f'FOREIGN KEY{plur}:')
        for fkey in self._fkeys.values():
            ret.append(repr(fkey))
        ret.append('')
        ret.append(fkeys_usage)
        for fkey in self._fkeys:
            ret.append(f"    '': '{fkey}',")
        ret.append('}')
    return '\n'.join(ret)

def is_set(self):
    """Return True if one field at least is set or if self has been
    constrained by at least one of its foreign keys or self is the
    result of a combination of Relations (using set operators).
    """
    joined_to = False
    for _, jt_ in self._joined_to.items():
        joined_to |= jt_.is_set()
    return (joined_to or bool(self.__set_operators.operator) or bool(self.__neg) or
            bool({field for field in self._fields.values() if field.is_set()}))

def __get_set_fields(self):
    """Returns a list containing only the fields that are set."""
    return [field for field in self._fields.values() if field.is_set()]

def __walk_op(self, rel_id_, out=None, _fields_=None):
    """Walk the set operators tree and return a list of SQL where
    representation of the query with a list of the fields of the query.
    """
    if out is None:
        out = []
        _fields_ = []
    if self.__set_operators.operator:
        if self.__neg:
            out.append("not (")
        out.append("(")
        left = self.__set_operators.left
        left.__query_type = self.__query_type
        left.__walk_op(rel_id_, out, _fields_)
        if self.__set_operators.right is not None:
            out.append(f" {self.__set_operators.operator}\n    ")
            right = self.__set_operators.right
            right.__query_type = self.__query_type
            right.__walk_op(rel_id_, out, _fields_)
        out.append(")")
        if self.__neg:
            out.append(")")
    else:
        out.append(self.__where_repr(rel_id_))
        _fields_ += self.__get_set_fields()
    return out, _fields_

def __join(self, orig_rel, deja_vu):
    for fkey, fk_rel in self._joined_to.items():
        fk_rel.__query_type = orig_rel.__query_type
        fk_rel.__get_from(orig_rel, deja_vu)
        if fk_rel.id_ not in deja_vu:
            deja_vu[fk_rel.id_] = []
        elif (fk_rel, fkey) in deja_vu[fk_rel.id_] or fk_rel is orig_rel:
            #sys.stderr.write(f"déjà vu in from! {fk_rel._fqrn}\n")
            continue
        deja_vu[fk_rel.id_].append((fk_rel, fkey))
        if fk_rel.__set_operators.operator:
            fk_rel.__get_from(self.id_)
        _, where, values = fk_rel.__where_args()
        where = f" and\n    {where}"
        orig_rel.__sql_query.insert(1, f'\n  join {__sql_id(fk_rel)} on\n   ')
        orig_rel.__sql_query.insert(2, fkey._join_query(self))
        orig_rel.__sql_query.append(where)
        orig_rel.__sql_values += values

def __sql_id(self):
    """Returns the FQRN as alias for the sql query."""
    return f"{self._qrn} as r{self.id_}"

def __get_from(self, orig_rel=None, deja_vu=None):
    """Constructs the __sql_query and gets the __sql_values for self."""
    if deja_vu is None:
        orig_rel = self
        self.__sql_query = [__sql_id(self)]
        deja_vu = {self.id_:[(self, None)]}
    self.__join(orig_rel, deja_vu)

def __where_repr(self, rel_id_):
    where_repr = []
    for field in self.__get_set_fields():
        where_repr.append(field._where_repr(self.__query_type, rel_id_))
    where_repr = ' and\n    '.join(where_repr) or '1 = 1'
    ret = f"({where_repr})"
    if self.__neg:
        ret = f"not ({ret})"
    return ret

def __where_args(self, *args):
    """Returns the what, where and values needed to construct the queries.
    """
    rel_id_ = self.id_
    what = f'r{rel_id_}.*'
    if args:
        what = ', '.join([f'r{rel_id_}.{arg}' for arg in args])
    s_where, set_fields = self.__walk_op(rel_id_)
    s_where = ''.join(s_where)
    if s_where == '()':
        s_where = '(1 = 1)'
    return what, s_where, set_fields

def __get_query(self, query_template, *args):
    """Prepare the SQL query to be executed."""
    from half_orm.fkey import FKey

    self.__sql_values = []
    self.__query_type = 'select'
    what, where, values = self.__where_args(*args)
    where = f"\nwhere\n    {where}"
    self.__get_from()
    # remove duplicates
    for idx, elt in reversed(list(enumerate(self.__sql_query))):
        if elt.find('\n  join ') == 0 and self.__sql_query.count(elt) > 1:
            self.__sql_query[idx] = '  and\n'
    # check that fkeys are fkeys
    for fkey_name in self._fkeys_attr:
        fkey_cls = self.__class__.__base__.__dict__[fkey_name].__class__
        if fkey_cls != FKey:
            raise RuntimeError(
                f'self.{fkey_name} is not a FKey (got a {fkey_cls.__name__} object instead).\n'
                f'- use: self.{fkey_name}.set({fkey_cls.__name__}(...))\n'
                f'- not: self.{fkey_name} = {fkey_cls.__name__}(...)'
                )
    return (
        query_template.format(
            what,
            self.__only and "only" or "",
            ' '.join(self.__sql_query), where),
        values)

def _prep_select(self, *args):
    self.__sql_values = []
    query_template = f"select\n {self.__select_params.get('distinct', '')} {{}}\nfrom\n  {{}} {{}}\n  {{}}"
    query, values = self.__get_query(query_template, *args)
    values = tuple(self.__sql_values + values)
    if 'order_by' in self.__select_params.keys():
        query = f"{query} order by {self.__select_params['order_by']}"
    if 'limit' in self.__select_params.keys():
        query = f"{query} limit {self.__select_params['limit']}"
    if 'offset' in self.__select_params.keys():
        query = f"{query} offset {self.__select_params['offset']}"
    return query, values

def distinct(self):
    """Set distinct in SQL select request."""
    self.__select_params['distinct'] = 'distinct'
    return self

def unaccent(self, *fields_names):
    "Sets unaccent for each field listed in fields_names"
    for field_name in fields_names:
        if not isinstance(self.__dict__[field_name], Field):
            raise ValueError(f'{field_name} is not a Field!')
        self.__dict__[field_name].unaccent = True
    return self

def order_by(self, _order_):
    """Set SQL order by according to the "order" string passed

    @order string example :
    "field1, field2 desc, field3, field4 desc"
    """
    self.__select_params['order_by'] = _order_
    return self

def limit(self, _limit_):
    """Set limit for the next SQL select request."""
    if _limit_:
        self.__select_params['limit'] = _limit_
    elif 'limit' in self.__select_params:
        self.__select_params.pop('limit')
    return self

def offset(self, _offset_):
    """Set the offset for the next SQL select request."""
    self.__select_params['offset'] = _offset_
    return self

def _mogrify(self):
    """Prints the select query."""
    self.__mogrify = True
    return self

def __len__(self):
    """Returns the number of tuples matching the intention in the relation.

    See select for arguments.
    """
    self.__query = "select"
    query_template = "select\n  count(distinct {})\nfrom {}\n  {}\n  {}"
    query, values = self.__get_query(query_template)
    try:
        vars_ = tuple(self.__sql_values + values)
        self.__execute(query, vars_)
    except Exception as err:
        self._mogrify()
        self.__execute(query, vars_)
        raise Exception from err
    return self.__cursor.fetchone()['count']

def is_empty(self):
    """Returns True if the relation is empty, False otherwise.

    Same as __len__ but limits the request to 1 element (faster).
    Use it instead of len(relation) == 0.
    """
    self.__query = "select"
    query_template = "select\n  count(distinct {})\nfrom {}\n  {}\n  {} limit 1"
    query, values = self.__get_query(query_template)
    try:
        vars_ = tuple(self.__sql_values + values)
        self.__execute(query, vars_)
    except Exception as err:
        print(query, vars_)
        raise err
    return self.__cursor.fetchone()['count'] != 1

def __count(self, *args, _distinct=False):
    """Returns the number of tuples matching the intention in the relation.

    See select for arguments.
    """
    self.__query = "select"
    if _distinct:
        query_template = "select\n  count(distinct {})\nfrom {}\n  {}\n  {}"
    else:
        query_template = "select\n  count({})\nfrom {}\n  {}\n  {}"
    query, values = self.__get_query(query_template, *args)
    try:
        vars_ = tuple(self.__sql_values + values)
        self.__execute(query, vars_)
    except Exception as err:
        self._mogrify()
        self.__execute(query, vars_)
        raise Exception from err
    return self.__cursor.fetchone()['count']

def __update_args(self, **kwargs):
    """Returns the what, where an values for the update query."""
    what_fields = []
    new_values = []
    self.__query_type = 'update'
    _, where, values = self.__where_args()
    where = f" where {where}"
    for field_name, new_value in kwargs.items():
        what_fields.append(field_name)
        new_values.append(new_value)
    what = ", ".join([f'"{elt}" = %s' for elt in what_fields])
    return what, where, new_values + values

def __what_to_insert(self):
    """Returns the field names and values to be inserted."""
    fields_names = []
    set_fields = self.__get_set_fields()
    if set_fields:
        fields_names = [
            f'"{name}"' for name, field in self._fields.items() if field.is_set()]
    fk_fields = []
    fk_queries = ''
    fk_values = []
    for fkey in self._fkeys.values():
        fk_prep_select = fkey._prep_select()
        if fk_prep_select is not None:
            fk_values += list(fkey.values()[0])
            fk_fields += fk_prep_select[0]
            fk_queries = ["%s" for _ in range(len(fk_values))]

    return fields_names, set_fields, fk_fields, fk_queries, fk_values

def __call__(self, **kwargs):
    return self.__class__(**kwargs)

def cast(self, qrn):
    """Cast a relation into another relation.
    """
    new = self._model._import_class(qrn)(**self._to_dict_val_comp())
    new.__id_cast = id(self)
    new._joined_to = self._joined_to
    new.__set_operators = self.__set_operators
    return new

def join(self, *f_rels):
    """Joins data to self.select() result. Returns a dict
    f_rels is a list of [(obj: Relation(), name: str, fields: Optional(<str|str[]>)), ...].

    Each obj in f_rels must have a direct or reverse fkey to self.
    If res is the result, res[name] contains the data associated to the element
    through the fkey or reversed fkey.
    If fields is a str, the data associated with res[name] is returned in a list (only one column).
    Otherwise (str[]), res[name] is a list of dict.
    If the fields argument is ommited, all the fields of obj are returned in a list of dict.

    Raises:
        RuntimeError: if self.__class__ and foreign.__class__ don't have fkeys to each other.

    Returns:
        dict: all values are converted to string.
    """
    from half_orm.fkey import FKey

    def to_str(value):
        """Returns value in string format if the type of value is
        in TO_PROCESS

        Args:
            value (any): the value to return in string format.
        """

        TO_PROCESS = {UUID, date, datetime, time, timedelta}
        if value.__class__ in TO_PROCESS:
            return str(value)
        return value

    res = list(
        {key: to_str(value) for key, value in elt.items()}
        for elt in self.distinct().select()
    )
    result_as_list = False
    ref = self()
    for f_rel in f_rels:
        if not isinstance(f_rel, tuple):
            raise RuntimeError("f_rels must be a list of tuples.")
        if len(f_rel) == 3:
            f_relation, name, fields = f_rel
        elif len(f_rel) == 2:
            f_relation, name = f_rel
            fields = list(f_relation._fields.keys())
        else:
            raise RuntimeError(f"f_rel must have 2 or 3 arguments. Got {len(f_rel)}.")
        if isinstance(fields, str):
            fields = [fields]
            result_as_list = True
        res_remote = {}

        f_relation_fk_names = []
        fkey_found = False
        for fkey_12 in ref._fkeys:
            if type(ref._fkeys[fkey_12]) != FKey:
                raise RuntimeError("This is not an Fkey")
            remote_fk = ref._fkeys[fkey_12]
            remote = remote_fk()
            if remote.__class__ == f_relation.__class__:
                for field in f_relation._fields.keys():
                    if f_relation.__dict__[field].is_set():
                        remote.__dict__[field]._set(f_relation.__dict__[field])
                fkey_found = True
                f_relation_fk_names = remote_fk.fk_names
                break

        if not fkey_found:
            raise RuntimeError(f"No foreign key between {self._fqrn} and {f_relation._fqrn}!")
        inter = [{key: to_str(val) for key, val in elt.items()}
            for elt in remote.distinct().select(
                *([f'"{field}"' for field in fields] + f_relation_fk_names))]
        for elt in inter:
            key = tuple(elt[subelt] for subelt in f_relation_fk_names)
            if key not in res_remote:
                res_remote[key] = []
            if result_as_list:
                res_remote[key].append(to_str(elt[fields[0]]))
            else:
                res_remote[key].append({key: to_str(elt[key]) for key in fields})

        if f_relation_fk_names:
            d_res = {
                tuple(elt[selt] for selt in remote_fk.names): elt
                for elt in res
                }
            to_remove = set()
            for elt in d_res:
                remote = res_remote.get(elt)
                if remote:
                    d_res[elt][name] = remote
                else:
                    to_remove.add(elt)
            res = [elt for elt in res if tuple(elt[selt]
                    for selt in remote_fk.names) not in to_remove]
    return res

def __set__op__(self, operator=None, right=None):
    """Si l'opérateur du self est déjà défini, il faut aller modifier
    l'opérateur du right ???
    On crée un nouvel objet sans contrainte et on a left et right et opérateur
    """
    def check_fk(new, jt_list):
        """Sets the _joined_to dictionary for the new relation.
        """
        for fkey, rel in jt_list.items():
            if rel is self:
                rel = new
            new._joined_to[fkey] = rel
    new = self(**self._to_dict_val_comp())
    new.__id_cast = self.__id_cast
    if operator:
        new.__set_operators.left = self
        new.__set_operators.operator = operator
    dct_join = self._joined_to
    if right is not None:
        new.__set_operators.right = right
        dct_join.update(right._joined_to)
    check_fk(new, dct_join)
    return new

def __and__(self, right):
    return self.__set__op__("and", right)
def __iand__(self, right):
    self = self & right
    return self

def __or__(self, right):
    return self.__set__op__("or", right)
def __ior__(self, right):
    self = self | right
    return self

def __sub__(self, right):
    return self.__set__op__("and not", right)
def __isub__(self, right):
    self = self - right
    return self

def __neg__(self):
    new = self.__set__op__(self.__set_operators.operator, self.__set_operators.right)
    new.__neg = not self.__neg
    return new

def __xor__(self, right):
    return (self | right) - (self & right)
def __ixor__(self, right):
    self = self ^ right
    return self

def __contains__(self, right):
    return len(right - self) == 0

def __eq__(self, right):
    if id(self) == id(right):
        return True
    return self in right and right in self

def __ne__(self, right):
    return not self == right

def __enter__(self):
    """Context management entry

    Returns self in a transaction context.

    Example usage:
    with relation as rel:
        rel.update(col=new_val)

    Equivalent to (in a transaction context):
    rel = relation.select()
    for elt in rel:
        new_elt = relation(**elt)
        new_elt.update(col=new_val)
    """
    @self.Transaction
    def context(self):
        return self
    return context(self)

def __exit__(_, *__):
    """Context management exit

    Not much to do here.
    """
    return False

def __iter__(self):
    return self.select()

def __next__(self):
    return next(self.select())

def singleton(fct):
    """Decorator. Enforces the relation to define a singleton.

    _is_singleton is set by Relation.get.
    _is_singleton is unset as soon as a Field is set.
    """
    @wraps(fct)
    def wrapper(self, *args, **kwargs):
        if self._is_singleton:
            return fct(self, *args, **kwargs)
        try:
            self = self.get()
            return fct(self, *args, **kwargs)
        except relation_errors.ExpectedOneError as err:
            raise relation_errors.NotASingletonError(err)
    return wrapper

def _debug(_):
    """For debug purpose"""

#### END of Relation methods definition

COMMON_INTERFACE = {
    '__init__': __init__,
    '_freeze': _freeze,
    '_unfreeze': _unfreeze,
    '__setattr__': __setattr__,
    '__execute': __execute,
    'id_': id_,
    '__set_fields': __set_fields,
    '__set_fkeys': __set_fkeys,
    'order_by': order_by,
    'limit': limit,
    'offset': offset,
    'distinct': distinct,
    'unaccent': unaccent,
    '__call__': __call__,
    'cast': cast,
    '__get_set_fields': __get_set_fields,
    '__repr__': __repr__,
    'only': only,
    'is_empty': is_empty,
    'group_by':group_by,
    'to_json': to_json,
    'to_dict': to_dict,
    '_to_dict_val_comp': _to_dict_val_comp,
    '__get_from': __get_from,
    '__get_query': __get_query,
    'is_set': is_set,
    '__where_repr': __where_repr,
    '__where_args': __where_args,
    '_prep_select': _prep_select,
    'select': select,
    '_mogrify': _mogrify,
    '__len__': __len__,
    # 'count': count,
    'get': get,
    'join': join,
    '__set__op__': __set__op__,
    '__and__': __and__,
    '__iand__': __iand__,
    '__or__': __or__,
    '__ior__': __ior__,
    '__sub__': __sub__,
    '__isub__': __isub__,
    '__xor__': __xor__,
    '__ixor__': __ixor__,
    '__neg__': __neg__,
    '__contains__': __contains__,
    '__eq__': __eq__,
    '__sql_id': __sql_id,
    '__walk_op': __walk_op,
    '__join': __join,
    'insert': insert,
    '__what_to_insert': __what_to_insert,
    'update': update,
    '__update_args': __update_args,
    'delete': delete,
    '__add_returning': __add_returning,
    'Transaction': Transaction,
    '_set_fkeys_properties': _set_fkeys_properties,
    '_set_fkey_property': _set_fkey_property,
    '__enter__': __enter__,
    '__exit__': __exit__,
    '__iter__': __iter__,
    '__next__': __next__,
    # test
    '_debug': _debug,
    'singleton': singleton,
}


TABLE_INTERFACE = COMMON_INTERFACE
VIEW_INTERFACE = COMMON_INTERFACE
MVIEW_INTERFACE = COMMON_INTERFACE
FDATA_INTERFACE = COMMON_INTERFACE

REL_INTERFACES = {
    'r': TABLE_INTERFACE,
    'p': TABLE_INTERFACE,
    'v': VIEW_INTERFACE,
    'm': MVIEW_INTERFACE,
    'f': FDATA_INTERFACE}

REL_CLASS_NAMES = {
    'r': 'Table',
    'p': 'Partioned table',
    'v': 'View',
    'm': 'Materialized view',
    'f': 'Foreign data'}



#-*- coding: utf-8 -*-
# pylint: disable=protected-access

"""This module provides the Field class. It is used by the `relation <#module-half_orm.relation>`_ module."""

import types
from psycopg2.extensions import register_adapter, adapt

from half_orm.null import NULL

class Field():
    """The class Field is for Relation internal usage. It is called by
    the RelationFactory metaclass for each field in the relation considered.
    """
    def __init__(self, name, relation, metadata):
        self.__relation = relation
        self.__name = name
        self.__is_set = False
        self.__metadata = metadata
        self.__value = None
        self.__unaccent = False
        self.__comp = '='

    @property
    def _relation(self):
        return self.__relation

    def is_set(self):
        "Returns if the field is set or not."
        return self.__is_set

    def _is_part_of_pk(self):
        "Returns True if the field is part of the PK"
        return bool(self.__metadata['pkey'])

    def is_not_null(self):
        "Returns True if the field is defined as not null."
        return bool(self.__metadata['notnull'])

    def __repr__(self):
        md_ = self.__metadata
        field_constraint = f"{md_['notnull'] and 'NOT NULL' or ''}"
        repr_ = f"({md_['fieldtype']}) {field_constraint}"
        if self.__is_set:
            repr_ = f"{repr_} ({self.__name} {self.__comp} {self.__value})"
        return repr_.strip()

    def __str__(self):
        return str(self.__value)

    def _praf(self, query, id_):
        """Returns field_name prefixed with relation alias if the query is
        select. Otherwise, returns the field name quoted with ".
        """
        id_ = f'r{id_}'
        if query == 'select':
            return f'{id_}."{self.__name}"'
        return f'"{self.__name}"'

    def _where_repr(self, query, id_):
        """Returns the SQL representation of the field for the where clause
        """
        where_repr = ''
        comp_str = '%s'
        comp = self._comp()
        if comp == '@@':
            comp_str = 'websearch_to_tsquery(%s)'
        if isinstance(self.__value, (list, tuple)):
            if self.type_[0] != '_': # not an array type
                comp_str = 'any(%s)'
                if comp == '@@':
                    comp_str = 'any(websearch_to_tsquery(%s))'
        if not self.unaccent:
            where_repr = f"{self._praf(query, id_)} {comp} {comp_str}"
        else:
            where_repr = f"unaccent({self._praf(query, id_)}) {comp} unaccent({comp_str})"
        return where_repr

    @property
    def value(self):
        "Returns the value of the field object"
        return self.__value

    def _set(self, *args):
        "Sets the value of the field object"
        self.__set__(self.__relation, *args)

    def __set__(self, obj, *args):
        """Sets the value (and the comparator) associated with the field."""
        self.__relation._is_singleton = False
        if len(args) == 1:
            value = args[0]
            if value is None:
                self._unset()
                return
            comp = None
            self.__relation = obj
            if isinstance(value, tuple):
                if len(value) != 2:
                    raise ValueError(f"Can't match {value} with (comp, value)!")
                comp, value = value
            if value is NULL and comp is None:
                comp = 'is'
            elif comp is None:
                comp = '='
        elif len(args) == 2:
            # The first argument IS comp.
            comp, value = args
            if value is None:
                raise ValueError("Can't have a None value with a comparator!")
        else:
            raise RuntimeError('')
        if value is NULL:
            if not comp in {'is', 'is not'}:
                raise ValueError("comp should be 'is' or 'is not' with NULL value!")
        self.__is_set = True
        self.__value = value
        self.__comp = comp

    def _unset(self):
        "Unset a field"
        self.__is_set = False
        self.__value = None
        self.__comp = '='

    @property
    def type_(self):
        "Returns the SQL type of the field"
        return self.__metadata['fieldtype']

    @property
    def unaccent(self):
        return self.__unaccent
    @unaccent.setter
    def unaccent(self, value):
        if not isinstance(value, bool):
            raise RuntimeError('unaccent argument must be of boolean type!')
        self.__unaccent = value

    def _comp(self):
        "Returns the comparator associated to the value."
        if self.__comp == '%':
            return '%%'
        return self.__comp

    @property
    def _relation(self):
        """Internal usage.

        Returns:
            Relation: The Relation class for which self is an attribute.
        """
        return self.__relation

    def _psycopg_adapter(self):
        """Return the SQL representation of self.__value"""
        return adapt(self.__value)

    @property
    def _name(self):
        return self.__name

    def __call__(self):
        """In case someone inadvertently uses the name of a field for a method."""
        rel_class = self.__relation.__class__
        rcn = rel_class.__name__
        method = None
        try:
            method = rel_class.__dict__[self.__name]
            # print('XXX', isinstance(method, types.FunctionType))
        except KeyError as err:
            # genuine attemp to call a Field.
            raise err
        err_msg = "'Field' object is not callable."
        warn_msg = f"'{self.__name}' is an attribute of type Field of the '{rcn}' object."
        err_msg = f"{err_msg}\nWARNING:        {warn_msg}"
        if method:
            err_msg = f"{err_msg}\n                Do not use '{self.__name}' as a method name."
        raise TypeError(err_msg)

register_adapter(Field, Field._psycopg_adapter)



#!/usr/bin/env python3
#-*- coding: utf-8 -*-

"""This module provides the class Model.

The class Model is responsible for the connection to the PostgreSQL database.
    
Once connected, you can use the
`get_relation_class <#half_orm.model.Model.get_relation_class>`_
method to generate a class to access any relation (table/view) in your database.

Example:
    >>> from half_orm.model import Model
    >>> model = Model('my_config_file')
    >>> class MyTable(model.get_relation_class('my_schema.my_table')):
    ...     # Your business code goes here

Note:
    The default schema is ``public`` in PostgreSQL, so to reference a table
    ``my_table`` in this schema you'll have to use ``pubic.my_table``.
"""


import os
import sys
from configparser import ConfigParser
from os import environ
import psycopg2
from psycopg2.extras import RealDictCursor
from typing import Generator, List

from half_orm.model_errors import MalformedConfigFile, MissingConfigFile, MissingSchemaInName, UnknownRelation
from half_orm.relation import Relation, REL_INTERFACES, REL_CLASS_NAMES
from half_orm import pg_meta

CONF_DIR = os.path.abspath(environ.get('HALFORM_CONF_DIR', '/etc/half_orm'))


psycopg2.extras.register_uuid()

class Model:
    """
    Parameters:
        config_file (str): the configuration file that contains the informations to connect
            to the database.
        scope (Optional[str]): used to agregate several modules in a package.
            See `half_orm_packager <https://github.com/collorg/halfORM_packager>`_.

    Note:
        The **config_file** is searched in the `HALFORM_CONF_DIR` variable if specified,
        then in `/etc/half_orm`. The file format is as follows:

            | [database]
            | name = <postgres db name>
            | user = <postgres user>
            | password = <postgres password>
            | host = <host name | localhost>
            | port = <port | 5432>

        *name* is the only mandatory entry if you are using an
        `ident login with a local account <https://www.postgresql.org/docs/current/auth-ident.html>`_.
    """
    __deja_vu = {}
    _classes_ = {}
    def __init__(self, config_file: str, scope: str=None):
        """Model constructor

        Use @config_file in your scripts. The @dbname parameter is
        reserved to the __factory metaclass.
        """
        self.__backend_pid = None
        self.__config_file = config_file
        self.__dbinfo = {}
        self.__dbname = None
        self.__conn = None
        self._scope = scope and scope.split('.')[0]
        self.__production = False
        self.__connect()

    def get_relation_class(self, relation_name: str) -> 'class(Relation)':
        """This method is a factory that generates a class that inherits the `Relation <#half_orm.relation.Relation>`_ class.

        Args:
            relation_name (string): the full name (`<schema>.<relation>`) of the targeted relation in the database.

        Raises:
            ValueError: if the schema is missing in relation_name
            UnknownRelationError: if the relation is not found in the database

        Returns:
            a class that inherits the `Relation <#half_orm.relation.Relation>`_ class:
                the class corresponding to the relation in the database.

        Examples:
            A class inheriting the `Relation <#half_orm.relation.Relation>`_ class is returned:
                >>> Person = model.get_relation_class('actor.person')
                >>> Person
                <class 'half_orm.relation.Table_HalftestActorPerson'>
                >>> Person.__bases__
                (<class 'half_orm.relation.Relation'>,)

            A prefered way to create a class:
                >>> class Person(model.get_relation_class('actor.person)):
                >>>     # Your code goes here

            A `MissingSchemaInName <#half_orm.model_errors.MissingSchemaInName>`_ is raised when the schema name is missing:
                >>> model.get_relation_class('person')
                [...]MissingSchemaInName: do you mean 'public.person'?

            An `UnknownRelation <#half_orm.model_errors.UnknownRelation>`_ is raised if the relation is not found in the model:
                >>> model.get_relation_class('public.person')
                [...]UnknownRelation: 'public.person' does not exist in the database halftest.
        """

        def factory(dct):
            """Function to build a Relation class corresponding to a PostgreSQL
            relation.
            """
            def _gen_class_name(rel_kind, sfqrn):
                """Generates class name from relation kind and FQRN tuple"""
                class_name = "".join([elt.capitalize() for elt in
                                    [elt.replace('.', '') for elt in sfqrn]])
                return f"{rel_kind}_{class_name}"

            bases = [Relation,]
            tbl_attr = {}
            tbl_attr['__base_classes'] = set()
            tbl_attr['__fkeys_properties'] = False
            tbl_attr['_qrn'] = pg_meta.normalize_qrn(dct['fqrn'])

            tbl_attr.update(dict(zip(['_dbname', '_schemaname', '_relationname'], dct['fqrn'])))
            if not tbl_attr['_dbname'] in Model._classes_:
                Model._classes_[tbl_attr['_dbname']] = {}
            if dct.get('model'):
                tbl_attr['_model'] = dct['model']
            else:
                tbl_attr['_model'] = Model._deja_vu(tbl_attr['_dbname'])
            rel_class = Model._check_deja_vu_class(*dct['fqrn'])
            if rel_class:
                return rel_class

            try:
                metadata = tbl_attr['_model']._relation_metadata(dct['fqrn'])
            except KeyError as exc:
                raise UnknownRelation(dct['fqrn']) from exc
            if metadata['inherits']:
                metadata['inherits'].sort()
                bases = []
            for parent_fqrn in metadata['inherits']:
                bases.append(factory({'fqrn': parent_fqrn}))
            tbl_attr['__metadata'] = metadata
            tbl_attr['_t_fqrn'] = dct['fqrn']
            tbl_attr['_fqrn'] = pg_meta.normalize_fqrn(dct['fqrn'])
            tbl_attr['__kind'] = REL_CLASS_NAMES[metadata['tablekind']]
            tbl_attr['_fkeys'] = []
            for fct_name, fct in REL_INTERFACES[metadata['tablekind']].items():
                tbl_attr[fct_name] = fct
            class_name = _gen_class_name(REL_CLASS_NAMES[metadata['tablekind']], dct['fqrn'])
            rel_class = type(class_name, tuple(bases), tbl_attr)
            Model._classes_[tbl_attr['_dbname']][dct['fqrn']] = rel_class
            return rel_class

        try:
            schema, table = relation_name.replace('"', '').rsplit('.', 1)
        except ValueError as err:
            raise MissingSchemaInName(relation_name) from err
        return factory({'fqrn': (self.__dbname, schema, table), 'model': self})


    @staticmethod
    def _deja_vu(dbname):
        """Returns None if the database hasn't been loaded yet.
        Otherwise, it returns the Model object already loaded.
        The Model object is shared between all_ the relations in the
        database. The Model object is loaded only once for a given database.
        """
        return Model.__deja_vu.get(dbname)

    def __connect(self, config_file: str=None, reload: bool=False):
        """Setup a new connection to the database.

        The reconnect method is an alias to the ``__connect`` method.

        Parameters:
            config_file (str): If a config_file is provided, the connection is made with the new
                parameters, allowing to change role. The database name must be the same.
            reload (bool): If set to True, reloads the metadata from the database. Usefull if
                the model has changed.

        Raises:
            MissingConfigFile: If the **config_file** is not found in *HALFORM_CONF_DIR*.
            MalformedConfigFile: if the *name* is missing in the **config_file**. 
            RuntimeError: If the reconnection is attempted on another database.
        """
        self.disconnect()
        if config_file:
            self.__config_file = config_file

        config = ConfigParser()

        if not config.read(
                [os.path.join(CONF_DIR, self.__config_file)]):
            raise MissingConfigFile(os.path.join(CONF_DIR, self.__config_file))

        params = dict(config['database'].items())
        if config_file and params['name'] != self.__dbname:
            raise RuntimeError(
                f"Can't reconnect to another database {params['name']} != {self.__dbname}")
        self.__dbname = params['name']
        self.__dbinfo['name'] = params['name']
        self.__dbinfo['user'] = params.get('user')
        self.__dbinfo['password'] = params.get('password')
        self.__dbinfo['host'] = params.get('host')
        self.__dbinfo['port'] = params.get('port')
        self.__production = params.get('production', False)
        if self.__production == 'True':
            self.__production = True
        if self.__production == 'False':
            self.__production = False
        if not isinstance(self.__production, bool):
            raise ValueError
        if 'name' not in self.__dbinfo:
            raise MalformedConfigFile(
                self.__config_file, {'name'})
        try:
            params = dict(self.__dbinfo)
            params['dbname'] = params.pop('name')
            self.__conn = psycopg2.connect(
                **params, cursor_factory=RealDictCursor)
        except psycopg2.OperationalError as err:
            raise err.__class__(err)
        self.__conn.autocommit = True
        self.__pg_meta = pg_meta.PgMeta(self.__conn, reload)
        self.__deja_vu[self.__dbname] = self
        self.__backend_pid = self.execute_query(
            "select pg_backend_pid()").fetchone()['pg_backend_pid']

    reconnect = __connect

    def ping(self):
        """Checks if the connection is still established.
        Attempts a new connection otherwise.

        Returns:
            bool: True if the connection is established, False otherwise.
        """
        try:
            self.execute_query("select 1")
            return True
        except (psycopg2.OperationalError, psycopg2.InterfaceError):
            try:
                self.__connect()
            except psycopg2.OperationalError as err:
                sys.stderr.write(f'{err}\n')
                sys.stderr.flush()
            return False

    def disconnect(self):
        """Closes the connection to the database.
        """
        if self.__conn is not None:
            if not self.__conn.closed:
                self.__conn.close()

    def _reload(self, config_file=None):
        "Reload metadata"
        self.__connect(config_file, True)

    @property
    def _pg_backend_pid(self):
        "backend PID"
        return self.__backend_pid

    @property
    def _dbname(self):
        """
        property. Returns the database name.
        """
        return self.__dbname

    @property
    def _connection(self):
        """\
        Property. Returns the psycopg2 connection attached to the Model object.
        """
        return self.__conn

    def _fields_metadata(self, sfqrn):
        "Proxy to PgMeta.fields_meta"
        return self.__pg_meta.fields_meta(self.__dbname, sfqrn)

    def _fkeys_metadata(self, sfqrn):
        "Proxy to PgMeta.fkeys_meta"
        return self.__pg_meta.fkeys_meta(self.__dbname, sfqrn)

    def _relation_metadata(self, fqrn):
        "Proxy to PgMeta.relation_meta"
        return self.__pg_meta.relation_meta(self.__dbname, fqrn)

    def _unique_constraints_list(self, fqrn):
        "Proxy to PgMeta._unique_constraints_list"
        return self.__pg_meta._unique_constraints_list(self.__dbname, fqrn)

    def _pkey_constraint(self, fqrn):
        "Proxy to PgMeta._pkey_constraint"
        return self.__pg_meta._pkey_constraint(self.__dbname, fqrn)

    def execute_query(self, query, values=()):
        """Executes a raw SQL query.
        
        Warning:
            This method calls the psycopg2
            `cursor.execute <https://www.psycopg.org/docs/cursor.html?highlight=execute#cursor.execute>`_
            function.
            Please read the psycopg2 documentation on
            `passing parameters to SQL queries <https://www.psycopg.org/docs/usage.html#query-parameters>`_.
        """
        cursor = self.__conn.cursor()
        cursor.execute(query, values)
        return cursor

    def execute_function(self, fct_name, *args, **kwargs) -> List[tuple]:
        """`Executes a PostgreSQL function <https://www.postgresql.org/docs/current/sql-syntax-calling-funcs.html>`_.

        Arguments:
            *args: The list of parameters to be passed to the postgresql function.
            **kwargs: The list of named parameters to be passed to the postgresql function.

        Returns:
            List[tuple]: a list of tuples.
        
        Raises:
            RuntimeError: If you mix ***args** and ****kwargs**.

        Note:
            You can't mix args and kwargs with the execute_function method!
        """
        if bool(args) and bool(kwargs):
            raise RuntimeError("You can't mix args and kwargs with the execute_function method!")
        cursor = self.__conn.cursor()
        if kwargs:
            values = kwargs
        else:
            values = args
        cursor.callproc(fct_name, values)
        return cursor.fetchall()

    def call_procedure(self, proc_name, *args, **kwargs):
        """`Executes a PostgreSQL procedure <https://www.postgresql.org/docs/current/sql-call.html>`_.

        Arguments:
            *args: The list of parameters to be passed to the postgresql function.
            **kwargs: The list of named parameters to be passed to the postgresql function.

        Returns:
            None | List[tuple]: None or a list of tuples.
        
        Raises:
            RuntimeError: If you mix ***args** and ****kwargs**.

        Note:
            You can't mix args and kwargs with the call_procedure method!
        """
        if bool(args) and bool(kwargs):
            raise RuntimeError("You can't mix args and kwargs with the call_procedure method!")
        if kwargs:
            params = ', '.join([f'{key} => %s' for key in kwargs])
            values = tuple(kwargs.values())
        else:
            params = ', '.join(['%s' for elt in range(len(args))])
            values = args
        query = f'call {proc_name}({params})'
        cursor = self.__conn.cursor()
        cursor.execute(query, values)
        try:
            return cursor.fetchall()
        except psycopg2.ProgrammingError:
            return None

    def has_relation(self, qtn: str) -> bool:
        """Checks if the qtn is a relation in the database

        Returns:
            bool: True if the relation exists in the database, False otherwise.

        Example:
            >>> model.has_relation('public.person')
            False
            >>> model.has_relation('actor.person')
            True
        """
        return self.__pg_meta.has_relation(self.__dbname, *qtn.rsplit('.', 1))

    @classmethod
    def _check_deja_vu_class(cls, dbname, schema, relation):
        """Not to use with _import_class.
        """
        if cls._classes_.get(dbname):
            return cls._classes_[dbname].get((dbname, schema, relation))

    def _import_class(self, qtn, scope=None):
        """Used to return the class from the scope module.

        This method is used to import a class from a module. The module
        must reside in an accessible python package named `scope`.
        """
        t_qtn = qtn.replace('"', '').rsplit('.', 1)
        module_path = f'{scope or self._scope}.{".".join(t_qtn)}'
        _class_name = pg_meta.class_name(qtn) # XXX
        module = __import__(
            module_path, globals(), locals(), [_class_name], 0)
        if scope:
            self._scope = scope
        return module.__dict__[_class_name]

    def _relations(self):
        """List all_ the relations in the database"""
        rels = self.__pg_meta.relations_list(self.__dbname)
        return rels

    def desc(self):
        """Returns the list of the relations of the model.

        Each element in the list contains:

        * the relation type: 'r' relation, 'v' view, 'm' materialized view, 'p' partition;
        * a tuple identifying the relation: (db name>, <schema name>, <relation name>);
        * a list of tuples indentifying the inherited relations.

        Example:
            >>> from half_orm.model import Model
            >>> halftest = Model('halftest')
            >>> halftest.desc()
            [('r', ('halftest', 'actor', 'person'), []), ('r', ('halftest', 'blog', 'comment'), []), ('r', ('halftest', 'blog', 'event'), [('halftest', 'blog', 'post')]), ('r', ('halftest', 'blog', 'post'), []), ('v', ('halftest', 'blog.view', 'post_comment'), [])]
        """
        return self.__pg_meta.desc(self.__dbname)

    def __str__(self):
        return self.__pg_meta.str(self.__dbname)

    @property
    def production(self):
        """Returns the production status of the db. Production is set by adding
        ``production = True`` to the **config file**.

        Returns:
            bool: True if the database is in production, False otherwise
        """
        return self.__production




#-*- coding: utf-8 -*-
# pylint: disable=too-few-public-methods, protected-access

"""This module provides the Transaction class."""

import sys

class Transaction:
    """The Transaction class is intended to be used as a class attribute of
    relation.Relation class:

    Relation.Transaction = Transaction

    The Relation.transaction can be used as a decorator of any method of a
    sub class of Relation class or any function receiving a Relation object
    as its first argument.

    Example of use:

    ```python
    gaston = halftest.relation("actor.person", first_name="Gaston")
    @gaston.Transaction
    def do_something(person):
        #... code to be done
    do_somethin(gaston)
    ```
    Every SQL commands executed in the do_something function will be put in a
    transaction and commited or rolled back at the end of the function.

    Functions decorated by a transaction can be nested:

    ```python
    @gaston.Transaction
    def second(gaston):
        # ... do something else
    @gaston.Transaction
    def first(gaston):
        # ... do something
        second(gaston)
    first(gaston)
    ```
    Here second is called by first and both function are played in the same
    transaction.
    """

    __level = 0
    def __init__(self, func):
        self.__func = func

    def __call__(self, relation, *args, **kwargs):
        """Each time a transaction is hit, the level is increased.
        The transaction is commited when the level is back to 0 after
        the return of the function.
        """
        res = None
        try:
            Transaction.__level += 1
            if relation._model._connection.autocommit:
                relation._model._connection.autocommit = False
            res = self.__func(relation, *args, **kwargs)
            Transaction.__level -= 1
            if Transaction.__level == 0:
                relation._model._connection.commit()
                relation._model._connection.autocommit = True
        except Exception as err:
            sys.stderr.write(f"Transaction error: {err}\nRolling back!\n")
            Transaction.__level = 0
            relation._model._connection.rollback()
            relation._model._connection.autocommit = True
            raise err
        return res



"""This module provides the errors for the model module."""

class MissingConfigFile(Exception):
    """The config file has not been found."""
    def __init__(self, filename):
        self.filename = filename
        Exception.__init__(self, f'Missing config file exception: {filename}')

class MalformedConfigFile(Exception):
    """The config file is malformed.

    The missing parameters are indicated in the error message.
    """
    def __init__(self, filename, missing_params):
        self.filename = filename
        missing_params_string = '\n - '.join([''] + list(missing_params))
        Exception.__init__(
            self,
            f"Malformed config file exception: {filename}\nMissing parameters: {missing_params_string}")

class UnknownDatabase(Exception):
    """The database dbname couldn't be found. Connexion error ?"""
    def __init__(self, dbname):
        self.dbname = dbname
        Exception.__init__(self, f'Unknown database exception: {dbname}')

class UnknownRelation(Exception):
    """The FQRN doesn't match any relation in the database."""
    def __init__(self, sfqrn):
        self.dbname = sfqrn[0]
        self.schema = sfqrn[1]
        self.relation = sfqrn[2]
        Exception.__init__(self, f"'{sfqrn[1]}.{sfqrn[2]}' does not exist in the database {sfqrn[0]}.")

class MissingSchemaInName(Exception):
    """The QRN should contain a schema name."""
    def __init__(self, qrn):
        Exception.__init__(self, f"do you mean 'public.{qrn}'?")



# -*- coding: utf-8 -*-
# pylint: disable=protected-access

"""This module provides the FKey class."""

from half_orm.pg_meta import normalize_fqrn, normalize_qrn

class FKey:
    """Foreign key class

    A foreign key is set by assigning to it a Relation object of the
    corresponding type (see FKey.set method).
    It is then used to construct the join query for Relation.select
    method.
    """

    def __init__(self,
                 fk_name, relation, fk_sfqrn,
                 fk_names=None, fields=None, confupdtype=None, confdeltype=None):
        self.__relation = relation
        self.__to_relation = None
        self.__name = fk_name
        self.__is_set = False
        self.__fk_names = fk_names or []
        self.__fk_from = None
        self.__fk_to = None
        self.__confupdtype = confupdtype
        self.__confdeltype = confdeltype
        self.__fk_fqrn = fk_sfqrn
        self.__fields_names = fields
        self.__fields = [f'"{name}"' for name in fields]

    def __call__(self, __cast__=None, **kwargs):
        """Returns the relation referenced by the fkey.
        If model._scope is set, instanciate the class from the scoped module.
        Uses the __cast if it is set.
        """
        model = self.__relation._model
        f_cast = None
        get_rel = model._import_class if model._scope else model.get_relation_class
        if self.__name.find('_reverse_fkey_') == 0 and __cast__:
            self.__relation = get_rel(__cast__)(**self.__relation.to_dict())
        else:
            f_cast = __cast__
        f_relation = get_rel(f_cast or normalize_qrn(self.__fk_fqrn))(**kwargs)
        rev_fkey_name = f'_reverse_{f_relation.id_}'
        f_relation._fkeys[rev_fkey_name] = FKey(
            rev_fkey_name,
            f_relation,
            f_relation._t_fqrn, self.__fields, self.__fk_names)
        f_relation._fkeys[rev_fkey_name].set(self.__relation)
        return f_relation

    def values(self):
        return [list(elt.values()) for elt in self.__to_relation.select(*self.__fk_names)]

    def set(self, to_):
        """Sets the relation associated to the foreign key."""
        from half_orm.relation import Relation

        self.__to_relation = to_
        from_ = self.__relation
        if not issubclass(to_.__class__, Relation):
            raise Exception("Expecting a Relation")
        to_classes = set(type.mro(to_.__class__))
        self_classes = set(type.mro(self.__relation.__class__))
        common_classes = to_classes.intersection(self_classes)
        if object in common_classes:
            common_classes.remove(object)
        if not common_classes:
            raise Exception(f"Type mismatch:\n{self.__fk_fqrn} != {to_._fqrn}")
        self.__fk_from = from_
        self.__fk_to = to_
        self.__is_set = to_.is_set()
        from_._joined_to[self] = to_

    @classmethod
    def __set__(cls, *args):
        print('XXX', cls, '\n', args)
        raise RuntimeError

    def is_set(self):
        """Return if the foreign key is set (boolean)."""
        return self.__is_set

    @property
    def to_(self):
        """Returns the destination relation of the fkey."""
        return self.__fk_to

    @to_.setter
    def to_(self, to_):
        """Sets the destination relation of the fkey."""
        self.__fk_to = to_

    @property
    def fk_fqrn(self):
        """Returns the FQRN of the relation pointed to."""
        return self.__fk_fqrn

    @property
    def confupdtype(self):
        "on update configuration"
        return self.__confupdtype

    @property
    def confdeltype(self):
        "on delete configuration"
        return self.__confdeltype

    def _join_query(self, orig_rel):
        """Returns the join_query, join_values of a foreign key.
        fkey interface: frel, from_, to_, fields, fk_names
        """
        from_ = self.__fk_from
        to_ = self.to_
        if id(from_) == id(to_):
            raise RuntimeError("You can't join a relation with itself!")
        orig_rel_id = f'r{orig_rel.id_}'
        to_id = f'r{to_.id_}'
        from_id = f'r{from_.id_}'
        if to_._qrn == orig_rel._qrn:
            to_id = orig_rel_id
        if from_._qrn == orig_rel._qrn:
            from_id = orig_rel_id
        from_fields = (f'{from_id}.{name}' for name in self.__fields)
        to_fields = (f'{to_id}.{name}' for name in self.fk_names)
        bounds = " and ".join(
            [f'{a} = {b}' for a, b in zip(to_fields, from_fields)])
        return f"({bounds})"

    def _prep_select(self):
        if self.__is_set:
            return self.__fields, self.to_._prep_select(*self.fk_names)
        return None

    @property
    def fk_names(self):
        """Returns the names of the fields composing the foreign key in the foreign table."""
        return self.__fk_names

    @fk_names.setter
    def fk_names(self, fk_names):
        """Sets the names of the fields in the foreign table."""
        self.__fk_names = fk_names

    @property
    def names(self):
        "Returns the names of the fields composing the foreign key in the table"
        return self.__fields_names

    def __repr__(self):
        """Representation of a foreign key
        """
        fields = list(self.__fields)
        fields.sort()
        fields = f"({', '.join(fields)})"
        repr_ = f"- {self.__name}: {fields}\n ↳ {normalize_fqrn(self.__fk_fqrn)}({', '.join(self.fk_names)})"
        if self.__is_set:
            repr_value = str(self.to_)
            res = []
            for line in repr_value.split('\n'):
                res.append(f'     {line}')
            res = '\n'.join(res)
            repr_ = f'{repr_}\n{res}'
        return repr_



"""This module provides the SQL request to extract the metadata of a
PostgreSQL database.
"""

from collections import OrderedDict

def strip_quotes(qrn):
    "Removes all double quotes from the qrn/fqrn"
    return qrn.replace('"', '')

def __get_qrn(fqrn: tuple) -> tuple:
    "Returns the qualified relation name <schema>.<relation> from the fully qualified relation name"
    return fqrn[1:]

def normalize_fqrn(t_fqrn: tuple) -> str:
    """
    Transform the tuple (<db name>, <schema name>, <table name>) in
    "<db name>":"<schema name>"."<table name>".
    Dots are allowed only in the schema name.
    """
    dbname, schemaname, tablename = t_fqrn
    return f'"{dbname}":"{schemaname}"."{tablename}"'

def normalize_qrn(t_qrn):
    """
    qrn is a tuple for the qualified relation name (<schema name>, <talbe name>)
    A schema name can have any number of dots in it.
    A table name can't have a dot in it.
    returns "<schema name>"."<relation name>"
    """
    return '.'.join([f'"{elt}"' for elt in __get_qrn(t_qrn)])

def camel_case(string):
    "Retruns the string transformed to camel case"
    ccname = []
    name = string.lower()
    capitalize = True
    for char in name:
        if not char.isalnum():
            capitalize = True
            continue
        if capitalize:
            ccname.append(char.upper())
            capitalize = False
            continue
        ccname.append(char)
    return ''.join(ccname)

def class_name(qrn):
    "Returns the class name from qrn"
    return camel_case(qrn.replace('"', '').split('.')[-1])

_REQUEST = """
SELECT
    a.attrelid AS tableid,
    array_agg( distinct i.inhseqno::TEXT || ':' || i.inhparent::TEXT ) AS inherits,
    c.relkind AS tablekind,
    n.nspname AS schemaname,
    c.relname AS relationname,
    tdesc.description AS tabledescription,
    a.attname AS fieldname,
    a.attnum AS fieldnum,
    adesc.description AS fielddescription,
    a.attndims AS fielddim,
    pt.typname AS fieldtype,
    NOT( a.attislocal ) AS inherited,
    cn_uniq.contype AS uniq,
    cn_uniq.conkey as pkeynum,
    a.attnotnull OR NULL AS notnull,
    cn_pk.contype AS pkey,
    cn_fk.contype AS fkey,
    cn_fk.conname AS fkeyname,
    cn_fk.conkey AS lfkeynum,
    cn_fk.confrelid AS fkeytableid,
    cn_fk.confkey AS fkeynum,
    cn_fk.confupdtype as fkey_confupdtype,
    cn_fk.confdeltype as fkey_confdeltype
FROM
    pg_class c -- table
    LEFT JOIN pg_description tdesc ON
    tdesc.objoid = c.oid AND
    tdesc.objsubid = 0
    LEFT JOIN pg_namespace n ON
    n.oid = c.relnamespace
    LEFT JOIN pg_inherits i ON
    i.inhrelid = c.oid
    LEFT JOIN pg_attribute a ON
    a.attrelid = c.oid
    LEFT JOIN pg_description adesc ON
    adesc.objoid = c.oid AND
    adesc.objsubid = a.attnum
    JOIN pg_type pt ON
    a.atttypid = pt.oid
    LEFT JOIN pg_constraint cn_uniq ON
    cn_uniq.contype = 'u' AND
    cn_uniq.conrelid = a.attrelid AND
    a.attnum = ANY( cn_uniq.conkey )
    LEFT JOIN pg_constraint cn_pk ON
    cn_pk.contype = 'p' AND
    cn_pk.conrelid = a.attrelid AND
    a.attnum = ANY( cn_pk.conkey )
    LEFT JOIN pg_constraint cn_fk ON
    cn_fk.contype = 'f' AND
    cn_fk.conrelid = a.attrelid AND
    a.attnum = ANY( cn_fk.conkey )
WHERE
    n.nspname <> 'pg_catalog'::name AND
    n.nspname <> 'information_schema'::name AND
    ( c.relkind = 'r'::"char" -- table
      OR c.relkind = 'v'::"char" -- view
      OR c.relkind = 'm' -- materialized view
      OR c.relkind = 'f' -- foreign table/view/mat. view
      OR c.relkind = 'p' -- patitioned table
    ) AND
    a.attnum > 0 -- AND
    AND (i.inhparent is null or i.inhparent not in (select oid from pg_class where relkind = 'p'))
    AND (cn_fk is null or cn_fk.confrelid not in (select inhrelid from pg_inherits where inhparent in (select oid from pg_class where relkind = 'p')))
GROUP BY
    a.attrelid,
    n.nspname,
    c.relname,
    tdesc.description,
    c.relkind,
    a.attnum,
    a.attname,
    a.attnum,
    adesc.description,
    a.attndims,
    a.attislocal,
    pt.typname,
    cn_uniq.contype,
    cn_uniq.conkey,
    a.attnotnull,
    cn_pk.contype,
    cn_fk.contype,
    cn_fk.conname,
    cn_fk.conkey,
    cn_fk.confrelid,
    cn_fk.confkey,
    cn_fk.confupdtype,
    cn_fk.confdeltype
ORDER BY
    n.nspname, c.relname, a.attnum
"""

class _Meta(dict):
    __d_meta = {}

    @classmethod
    def deja_vu(cls, dbname):
        return dbname in cls.__d_meta

    @classmethod
    def register(cls, dbname, meta):
        cls.__d_meta[dbname] = meta

    def __getitem__(self, key):
        return _Meta.__d_meta.__getitem__(key)

    def __setitem__(self, key, val):
        _Meta.__d_meta.__setitem__(key, val)

    def __repr__(self):
        dictrepr = dict.__repr__(self.__d_meta)
        return f'{type(self).__name__}({dictrepr})'
  
    def update(self, *args, **kwargs):
        for key, value in dict(*args, **kwargs).items():
            self.__d_meta[key] = value

class PgMeta:
    meta = _Meta()
    def __init__(self, connection, reload=False):
        self.__dbname = connection.get_dsn_parameters()['dbname']
        if not PgMeta.meta.deja_vu(self.__dbname) or reload:
            self.__load_metadata(connection)

    def metadata(self, dbname):
        return self.meta[dbname].__metadata

    def relations_list(self, dbname):
        return self.metadata(dbname)['relations_list']

    def __load_metadata(self, connection):
        """Loads the metadata by querying the _REQUEST.
        """
        metadata = {'relations_list': []}
        byname = metadata['byname'] = OrderedDict()
        byid = metadata['byid'] = {}
        with connection.cursor() as cur:
            cur.execute(_REQUEST)
            all_ = [elt for elt in cur.fetchall()]
            for dct in all_:
                table_key = (self.__dbname, dct['schemaname'], dct['relationname'])
                tableid = dct['tableid']
                description = dct['tabledescription']
                if table_key not in byname:
                    byid[tableid] = {}
                    byid[tableid]['sfqrn'] = table_key
                    byid[tableid]['fields'] = OrderedDict()
                    byid[tableid]['fkeys'] = OrderedDict()
                    byname[table_key] = OrderedDict()
                    byname[table_key]['description'] = description
                    byname[table_key]['fields'] = OrderedDict()
                    byname[table_key]['fkeys'] = OrderedDict()
                    byname[table_key]['fields_by_num'] = OrderedDict()
            for dct in all_:
                tableid = dct['tableid']
                table_key = byid[tableid]['sfqrn']
                fieldname = dct.pop('fieldname')
                fieldnum = dct['fieldnum']
                tablekind = dct.pop('tablekind')
                inherits = [byid[int(elt.split(':')[1])]['sfqrn']
                            for elt in dct.pop('inherits') if elt is not None]
                byname[table_key]['tableid'] = tableid
                byname[table_key]['tablekind'] = tablekind
                byname[table_key]['inherits'] = inherits
                byname[table_key]['fields'][fieldname] = dct
                byname[table_key]['fields_by_num'][fieldnum] = dct
                byid[tableid]['fields'][fieldnum] = fieldname
                if (tablekind, table_key) not in metadata['relations_list']:
                    metadata['relations_list'].append((tablekind, table_key))
            for dct in all_:
                tableid = dct['tableid']
                table_key = byid[tableid]['sfqrn']
                fkeyname = dct['fkeyname']
                if fkeyname and not fkeyname in byname[table_key]['fkeys']:
                    fkeytableid = dct['fkeytableid']
                    ftable_key = byid[fkeytableid]['sfqrn']
                    fields = [byid[tableid]['fields'][num] for num in dct['lfkeynum']]
                    confupdtype = dct['fkey_confupdtype']
                    confdeltype = dct['fkey_confdeltype']
                    ffields = [byid[fkeytableid]['fields'][num] for num in dct['fkeynum']]
                    rev_fkey_name = f'_reverse_fkey_{"_".join(table_key)}.{".".join(fields)}'
                    rev_fkey_name = strip_quotes(rev_fkey_name.replace(".", "_").replace(":", "_"))
                    byname[table_key]['fkeys'][fkeyname] = (
                        ftable_key, ffields, fields, confupdtype, confdeltype)
                    byname[ftable_key]['fkeys'][rev_fkey_name] = (table_key, fields, ffields)

        metadata['relations_list'].sort()
        self.__metadata = metadata
        PgMeta.meta.register(self.__dbname, self)

    def getFqrn(self, dbname, qrn):
        "Returns the Fully qualified relation name (quoted) from the unquoted (qrn)"
        schema, table = qrn.rsplit('.', 1)
        return f'"{dbname}":"{schema}"."{table}"'

    def has_relation(self, dbname, schema, relation):
        """Checks if the qrn is a relation in the database

        @qrn is in the form <schema>.<table>
        Returns True if the relation exists, False otherwise.
        Also works for views and materialized views.
        """
        return (dbname, schema, relation) in self.meta[dbname].__metadata['byname']


    def desc(self, dbname):
        """Returns the list of the relations of the model.

        Each line contains:
        - the relation type: 'r' relation, 'v' view, 'm' materialized view,
        - the quoted FQRN (Fully qualified relation name)
          <"db name">:"<schema name>"."<relation name>"
        - the list of the FQRN of the inherited relations.

        If a qualified relation name (<schema name>.<table name>) is
        passed, prints only the description of the corresponding relation.
        """
        ret_val = []
        entry = self.metadata(dbname)['byname']
        for key in entry:
            if key[1].find('half_orm_meta') == 0: continue
            inh = []
            tablekind = entry[key]['tablekind']
            if entry[key]['inherits']:
                inh = [elt for elt in entry[key]['inherits']]
            ret_val.append((tablekind, key, inh))
        return ret_val

    def fields_meta(self, dbname, sfqrn):
        "Retruns the fields metadata for a given sfqrn"
        return self.metadata(dbname)['byname'][sfqrn]['fields']

    def fkeys_meta(self, dbname, sfqrn):
        "Returns the foreign keys metadata for a given sfqrn"
        return self.metadata(dbname)['byname'][sfqrn]['fkeys']

    def relation_meta(self, dbname, fqrn):
        "Returns the relation metadata for a given fqrn"
        return self.metadata(dbname)['byname'][fqrn]

    def str(self, dbname):
        out = []
        entry = self.metadata(dbname)['byname']
        for key in entry:
            if key[1].find('half_orm_meta') == 0: continue
            out.append(f"{entry[key]['tablekind']} {normalize_qrn(key)}")
        return '\n'.join(out)

    def _unique_constraints_list(self, dbname, sfqrn):
        "Returns the unique constraints of the given sfqrn"
        rel_meta_by_name = self.metadata(dbname)['byname'][sfqrn]
        tableid = rel_meta_by_name['tableid']
        rel_meta_by_id = self.metadata(dbname)['byid']
        unique_by_num = []
        for key, value in rel_meta_by_name['fields'].items():
            if value['uniq']:
                t_uniq = tuple(value['pkeynum'])
                if unique_by_num.count(t_uniq) == 0:
                    unique_by_num.append(t_uniq)
        unique = []
        for elt in unique_by_num:
            fields_names = []
            for num in elt:
                fields_names.append(rel_meta_by_id[tableid]['fields'][num])
            unique.append(tuple(fields_names))
        return unique

    def _pkey_constraint(self, dbname, sfqrn):
        "Returns the pkey constraint"
        rel_meta_by_name = self.metadata(dbname)['byname'][sfqrn]
        tableid = rel_meta_by_name['tableid']
        rel_meta_by_id = self.metadata(dbname)['byid']
        pkey_by_num = []
        for key, value in rel_meta_by_name['fields'].items():
            if value['pkey']:
                pkey_by_num.append(value['fieldnum'])
        pkey = []
        for num in pkey_by_num:
            pkey.append(rel_meta_by_id[tableid]['fields'][num])
        return pkey



import unittest
from typing import List
from half_orm.relation import Relation

class HoTestCase(unittest.TestCase):
    def hotAssertIsPkey(self, relation: Relation, field_names: List[str]):
        "it shoud be the primary key"
        pkey = set(relation()._pkey.keys())
        if set(field_names) != pkey:
            raise self.fail(f'PKey failure: {set(field_names)} != {pkey}')

    def hotAssertIsUnique(self, relation: Relation, fields_names: List[str]):
        "Checks if a list of fields is unique for the relation."
        fields_nums = set()
        pkeynum = []
        for field_name in fields_names:
            field = relation()._fields[field_name]
            metadata = field._Field__metadata
            fields_nums.add(metadata['fieldnum'])
            if not pkeynum and metadata['pkeynum']:
                pkeynum = metadata['pkeynum']
            if not metadata['uniq'] and pkeynum != metadata['pkeynum']:
                raise self.fail(f"'{fields_names}' is not unique.")
        if fields_nums != set(pkeynum):
            raise self.fail(f"'{fields_names}' is not unique.")

    def hotAssertIsNotNull(self, relation: Relation, field_name: str):
        if not relation()._fields[field_name].is_not_null():
            raise self.fail(f"'{field_name}' is not 'not null'.")
        
    def hotAssertReferences(self, relation: Relation, fk_name: str, f_relation: Relation):
        referenced = relation()._fkeys[fk_name]()
        if not referenced._qrn == f_relation._qrn:
            raise self.fail(f"{relation}()._fkeys['{fk_name}']() does not reference {f_relation}")

    def hotAssertAliasReferences(self, relation: Relation, alias: str, f_relation: Relation):
        referenced = eval(f"relation().{alias}")
        if not referenced()._qrn == f_relation._qrn:
            raise self.fail(f"{relation}.{alias}() does not reference {f_relation}")

    def __check_update_action(self, relation: Relation, fk_name: str, update_type: str):
        return self.assertEqual(relation()._fkeys[fk_name].confupdtype, update_type)

    def __check_delete_action(self, relation: Relation, fk_name: str, delete_type: str):
        return self.assertEqual(relation()._fkeys[fk_name].confdeltype, delete_type)

    def hotAssertOnUpdateNoAction(self, relation: Relation, fk_name: str):
        return self.__check_update_action(relation, fk_name, 'a')

    def hotAssertOnUpdateRestict(self, relation: Relation, fk_name: str):
        return self.__check_update_action(relation, fk_name, 'r')

    def hotAssertOnUpdateCascade(self, relation: Relation, fk_name: str):
        return self.__check_update_action(relation, fk_name, 'c')

    def hotAssertOnUpdateSetNull(self, relation: Relation, fk_name: str):
        return self.__check_update_action(relation, fk_name, 'n')

    def hotAssertOnUpdateSetDefault(self, relation: Relation, fk_name: str):
        return self.__check_update_action(relation, fk_name, 'd')

    def hotAssertOnDeleteNoAction(self, relation: Relation, fk_name: str):
        return self.__check_delete_action(relation, fk_name, 'a')

    def hotAssertOnDeleteRestict(self, relation: Relation, fk_name: str):
        return self.__check_delete_action(relation, fk_name, 'r')

    def hotAssertOnDeleteCascade(self, relation: Relation, fk_name: str):
        return self.__check_delete_action(relation, fk_name, 'c')

    def hotAssertOnDeleteSetNull(self, relation: Relation, fk_name: str):
        return self.__check_delete_action(relation, fk_name, 'n')

    def hotAssertOnDeleteSetDefault(self, relation: Relation, fk_name: str):
        return self.__check_delete_action(relation, fk_name, 'd')




"""This module provides the errors for the relation module."""

class ExpectedOneError(Exception):
    """This exception is raised when get count differs from 1."""
    def __init__(self, relation, count):
        self.rel = relation
        self.count = count
        self.plural = '' if count == 0 else 's'
        Exception.__init__(self, f'Expected 1, got {self.count} tuple{self.plural}')

class UnknownAttributeError(Exception):
    """Unknown attribute error"""
    def __init__(self, msg):
        super().__init__(f"ERROR! Unknown attribute: {msg}.")

class IsFrozenError(Exception):
    """Class is frozen"""
    def __init__(self, cls, msg):
        super().__init__(
            f"ERROR! The class {cls} is forzen.\n"
            f"Use _unfreeze to add the '{msg}' attribute to it.")

class DuplicateAttributeError(Exception):
    """Attempt to setattr to an already existing attribute."""

class NotASingletonError(Exception):
    """The constraint do not define a singleton.
    
    Raised from ExpectedOneError (err).
    """
    def __init__(self, err):
        Exception.__init__(self, f'Not a singleton. Got {err.count} tuple{err.plural}')

class WrongFkeyError(Exception):
    "Raised when Fkeys contains a wrong name"
    def __init__(self, cls, value):
        fkeys_list = "\n".join([f" - {fkey}" for fkey in cls._fkeys.keys()])
        err = f"Can't find '{value}'!\n" \
            f"List of keys for {cls.__class__.__name__}:\n" \
            f"{fkeys_list}"
        super().__init__(err)



#-*- coding: utf-8 -*-
# pylint: disable=too-few-public-methods

"""The null module provides the Null class
The Null class is used to set NULL value to relation fields.
"""

from psycopg2.extensions import register_adapter, AsIs

__all__ = ['NULL']

class Null:
    """The Null class"""

def adapt_null(_):
    """NULL adapter"""
    return AsIs("NULL")

register_adapter(Null, adapt_null)

NULL = Null()



#-*- coding: utf-8 -*-

import os
import codecs
from setuptools import setup

def read(name):
    file_name = os.path.join(os.path.abspath(os.path.dirname(__file__)), name)
    return codecs.open(file_name, "r", "utf-8").read()

setup(
    name='half_orm',
    version=read('half_orm/version.txt').strip(),
    description="A simple PostgreSQL to Python mapper.",
    long_description=read('README.md'),
    keywords='',
    author='Joël Maïzi',
    author_email='joel.maizi@collorg.org',
    url='https://github.com/collorg/halfORM',
    license='GNU General Public License v3 (GPLv3)',
    packages=['half_orm'],
    install_requires=[
        'psycopg2-binary',
        'PyYAML'],
    package_data={'half_orm': ['version.txt']},
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 4 - Beta',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
    ],
    long_description_content_type = "text/markdown"

)
